# coding=utf-8

"""shortipy module."""

from typing import Final
from os import linesep, path, makedirs

from flask import Flask

from shortipy.services.config import Config
from shortipy.services.redis import init_app as init_redis
from shortipy.services.hash import init_app as init_hash
from shortipy.services.auth import init_app as init_auth
from shortipy.services.serialization import init_app as init_serialization
from shortipy.services.url import init_app as init_url
from shortipy.controllers.resolution import resolution_blueprint
from shortipy.controllers.api import init_app as init_api

VERSION: Final = '1.3.0'
CONFIG_FILENAME: Final = 'config.py'


def create_app(options: dict | None = None) -> Flask | None:
    """Application factory.

    :param options: Optional application options (default: None).
    :type: dict | None
    :return: An application instance.
    :rtype: Flask | None
    """
    app = Flask(__name__, instance_relative_config=True)

    app.config.from_object(Config())

    if not isinstance(options, dict | None):
        raise Exception('Invalid options')
    if (options is None) or (not options.get('TESTING')):
        if not path.isdir(app.instance_path):
            makedirs(app.instance_path)
            raise Exception(f'Directory not found, so just created: {app.instance_path}.{linesep}'
                            f'Put file "{CONFIG_FILENAME}" inside, please.')

        config_file = path.join(app.instance_path, CONFIG_FILENAME)
        if not path.isfile(config_file):
            raise Exception(f'Configuration file not found: {config_file}')
        app.config.from_pyfile(config_file)

        if app.config.get('SECRET_KEY') is None:
            raise Exception(f'Set variable SECRET_KEY with random string in file: {config_file}')

    if options is not None:
        app.config.from_mapping(options)

    init_url(init_serialization(init_auth(init_hash(init_redis(app)))))

    app.register_blueprint(resolution_blueprint)
    app.register_blueprint(init_api())

    @app.cli.command('version', help='Display version.')
    def version():
        """Print Shortipy version."""
        print(f'Shortipy v{VERSION}')

    return app



# coding=utf-8

"""shortipy.services.auth file."""

from typing import Final

from click import option, STRING
from flask import Flask
from flask.cli import AppGroup
from flask_jwt_extended import JWTManager, create_access_token
from werkzeug.exceptions import NotFound, Unauthorized

from shortipy.services.redis import redis_client
from shortipy.services.hash import bcrypt, normalize_input

USER_KEYS_DOMAIN: Final = 'user'

jwt = JWTManager()
cli = AppGroup('users', help='Manage users.')


def init_app(app: Flask) -> Flask:
    """Initializes the application auth.

    :param app: The Flask application instance.
    :type app: Flask
    :return: The Flask application instance.
    :rtype: Flask
    """
    jwt.init_app(app)
    app.cli.add_command(cli)
    return app


def insert_user(username: str, password: str | bytes):
    """Insert new user.

    :param username: User's username.
    :type username: str
    :param password: User's password.
    :type password: str | bytes
    """
    if redis_client.hget(f'{USER_KEYS_DOMAIN}:{username}', 'password') is not None:
        raise Exception(f'User "{username}" already exists')

    password_hash = bcrypt.generate_password_hash(normalize_input(password))

    redis_client.hset(f'{USER_KEYS_DOMAIN}:{username}', 'password', password_hash)


def delete_user(username: str):
    """Delete user.

    :param username: User's username.
    :type username: str
    """
    if not redis_client.hgetall(f'{USER_KEYS_DOMAIN}:{username}'):
        raise NotFound(f'User "{username}" not found')

    redis_client.delete(f'{USER_KEYS_DOMAIN}:{username}')


def login(username: str, password: str | bytes) -> str:
    """Do login.

    :param username: User's username.
    :type username: str
    :param password: User's password.
    :type password: str | bytes
    :return: Access token.
    :rtype: str
    """
    if not bcrypt.check_password_hash(
            redis_client.hget(f'{USER_KEYS_DOMAIN}:{username}', 'password'), normalize_input(password)
    ):
        raise Unauthorized('Bad username or password')

    access_token = create_access_token(identity=username)
    return access_token


# region CLI functions
@cli.command('new', help='Insert new user.')
@option('-u', '--username', type=STRING, prompt='Enter the user\'s username', help='Specify the user\'s username.')
@option('-p', '--password', type=STRING, prompt='Enter the user\'s password', help='Specify the user\'s password.')
def new_user(username: str, password: str):
    """Insert new user.

    :param username: User's username.
    :type username: str
    :param password: User's password.
    :type password: str
    """
    print(f'Insert user: {username}...')
    insert_user(username, password)
    print('Done.')


@cli.command('del', help='Delete user by username.')
@option('-u', '--username', type=STRING, prompt='Enter the user\'s username', help='Specify the user\'s username.')
def del_url(username: str):
    """Delete user by username.

    :param username: User's username.
    :type username: str
    """
    print(f'Deleting user: {username}...')
    delete_user(username)
    print('Done.')
# endregion



# coding=utf-8

"""shortipy.services.exceptions file."""

from werkzeug.exceptions import HTTPException


class MethodVersionNotFound(HTTPException):
    """*400* `Method Version Not Found` (Bad Request)
    Raise if the browser request method through an invalid method version ('Accept-Version').
    (Raise if the browser sends something to the application or server cannot handle).
    """

    code = 400
    description = (
        "Bad Request - Method Version Not Found: The browser (or proxy) request an invalid method's version."
    )



# coding=utf-8

"""shortipy.controllers.api.url file."""

from flask import Flask, Blueprint, request, abort
from flask.views import MethodView
from flask_jwt_extended import jwt_required
from webargs import fields
from webargs.flaskparser import use_args
from marshmallow.validate import Length

from shortipy.services.exceptions import MethodVersionNotFound
from shortipy.services.serialization import marshmallow
from shortipy.services.url import get_urls, get_url_value, insert_url, update_url, delete_url


class UrlSchema(marshmallow.Schema):
    """Class to define url schema."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class Meta."""

        ordered = True
        fields = ('key', 'value', 'links')

    links = marshmallow.Hyperlinks(  # pylint: disable=no-member
        {
            'self': marshmallow.URLFor('api.url', values=dict(key='<key>'))  # pylint: disable=no-member
        }
    )


class UrlListAPI(MethodView):
    """Urls list API."""

    init_every_request = False

    def __init__(self):
        """UrlListAPI constructor."""
        self.url_schema = UrlSchema()
        self.urls_schema = UrlSchema(many=True)

    @jwt_required()
    def get(self):
        """Get urls.

        :return: All urls.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            urls = get_urls()
            if len(urls) < 1:
                abort(404)
            return {'urls': self.urls_schema.dump([{'key': key, 'value': value} for key, value in urls.items()])}
        raise MethodVersionNotFound()

    @jwt_required()
    @use_args({'value': fields.Str(required=True, validate=Length(min=1))}, location='json')
    def post(self, args: dict):
        """Post url.

        :param args: Arguments.
        :type args: dict
        :return: Inserted url.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            return {'url': self.url_schema.dump({'key': insert_url(args['value']), 'value': args['value']})}, 201
        raise MethodVersionNotFound()


class UrlAPI(MethodView):
    """Url API."""

    init_every_request = False

    def __init__(self):
        """UrlAPI constructor."""
        self.url_schema = UrlSchema()

    @jwt_required()
    def get(self, key: str):
        """Get url.

        :param key: Url key.
        :type key: str
        :return: Url.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            value = get_url_value(key)
            if value is None:
                abort(404)
            return {'url': self.url_schema.dump({'key': key, 'value': value})}
        raise MethodVersionNotFound()

    @jwt_required()
    @use_args({'value': fields.Str(required=True, validate=Length(min=1))}, location='json')
    def put(self, args: dict, key: str):
        """Put url.

        :param args: Arguments.
        :type args: dict
        :param key: Url key.
        :type key: str
        :return: Updated url.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            return {'url': self.url_schema.dump({'key': key, 'value': update_url(key, args['value'])})}
        raise MethodVersionNotFound()

    @staticmethod
    @jwt_required()
    def delete(key: str):
        """Delete url.

        :param key: Url key.
        :type key: str
        :return: Url.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            delete_url(key)
            return '', 204
        raise MethodVersionNotFound()


def register_api(app: Flask | Blueprint) -> Flask | Blueprint:
    """Register API controller.

    :param app: The Flask (or Blueprint) application instance.
    :type app: Flask | Blueprint
    :return: The Flask (or Blueprint) application instance.
    :rtype: Flask | Blueprint
    """
    app.add_url_rule('/urls/', view_func=UrlListAPI.as_view('urls'))
    app.add_url_rule('/urls/<key>', view_func=UrlAPI.as_view('url'))
    return app



# coding=utf-8

"""shortipy.services.serialization file."""

from flask import Flask
from flask_marshmallow import Marshmallow

marshmallow = Marshmallow()


def init_app(app: Flask) -> Flask:
    """Initializes the application serialization.

    :param app: The Flask application instance.
    :type app: Flask
    :return: The Flask application instance.
    :rtype: Flask
    """
    marshmallow.init_app(app)
    return app



# coding=utf-8

"""shortipy.controllers.resolution file."""

from flask import Blueprint, Response, redirect, abort
from markupsafe import escape

from shortipy.services.redis import redis_client

resolution_blueprint = Blueprint('resolution', __name__)


@resolution_blueprint.route('/<key>')
def resolve(key: str) -> Response:
    """Resolve the key in the correspondent url and redirect.

    :param key: Key to resolve.
    :type key: str
    :return: Flask response.
    :rtype: Response
    """
    value = redis_client.get(escape(key))
    if value is None:
        abort(404)
    return redirect(value)



# coding=utf-8

"""shortipy.controllers.api.auth file."""

from flask import Flask, Blueprint, request
from flask.views import MethodView
from webargs import fields
from webargs.flaskparser import use_args
from marshmallow.validate import Length

from shortipy.services.exceptions import MethodVersionNotFound
from shortipy.services.serialization import marshmallow
from shortipy.services.auth import login


class AuthSchema(marshmallow.Schema):
    """Class to define auth schema."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class Meta."""

        ordered = True
        fields = ('username', 'access_token')


class AuthListAPI(MethodView):
    """Auth list API."""

    init_every_request = False

    def __init__(self):
        """AuthListAPI constructor."""
        self.auth_schema = AuthSchema()

    @use_args({
        'username': fields.Str(required=True, validate=Length(min=1)),
        'password': fields.Str(required=True, validate=Length(min=1))
    }, location='json')
    def post(self, args: dict):
        """Post login.

        :param args: Arguments.
        :type args: dict
        :return: Inserted url.
        :rtype: dict[str, str]
        """
        if request.headers.get('Accept-Version', '1.0') == '1.0':
            return {'auth': self.auth_schema.dump({
                'username': args['username'],
                'access_token': login(args['username'], args['password'])
            })}
        raise MethodVersionNotFound()


def register_api(app: Flask | Blueprint) -> Flask | Blueprint:
    """Register API controller.

    :param app: The Flask (or Blueprint) application instance.
    :type app: Flask | Blueprint
    :return: The Flask (or Blueprint) application instance.
    :rtype: Flask | Blueprint
    """
    app.add_url_rule('/auth/', view_func=AuthListAPI.as_view('auth'))
    return app



# coding=utf-8

"""shortipy.controllers.api module."""

from flask import Flask, Blueprint

from shortipy.controllers.api.auth import register_api as register_api_auth
from shortipy.controllers.api.url import register_api as register_api_url


def init_app() -> Flask | Blueprint:
    """Initializes the application API.

    :return: The Flask (or Blueprint) application instance.
    :rtype: Flask | Blueprint
    """
    api_blueprint = Blueprint('api', __name__, url_prefix='/api')
    return register_api_url(register_api_auth(api_blueprint))



# coding=utf-8

"""shortipy.services.url file."""

from typing import Final
from string import ascii_lowercase
from os import linesep
from random import SystemRandom

from click import STRING, option
from flask import Flask
from flask.cli import AppGroup
from werkzeug.exceptions import NotFound

from shortipy.services.redis import redis_client

URL_KEYS_DOMAIN: Final = 'url'

cli = AppGroup('urls', help='Manage urls.')


def init_app(app: Flask) -> Flask:
    """Initializes the application urls.

    :param app: The Flask application instance.
    :type app: Flask
    :return: The Flask application instance.
    :rtype: Flask
    """
    app.cli.add_command(cli)
    return app


# region CRUD functions
def get_urls() -> dict[str, str]:
    """Get urls.

    :return: Dictionary of urls (keys and values).
    :rtype: dict[str, str]
    """
    return {key.removeprefix(f'{URL_KEYS_DOMAIN}:'): redis_client.get(key)
            for key in redis_client.keys(f'{URL_KEYS_DOMAIN}:*')}


def get_url_value(key: str) -> str | None:
    """Get url value by passed key.

    :param key: Key to find.
    :type key: str
    :return: Url value found or None.
    :rtype: str | None
    """
    return redis_client.get(f'{URL_KEYS_DOMAIN}:{key}')


def insert_url(value: str) -> str:
    """Insert passed url value and generate a key to retrieve it.

    :param value: Url value to insert.
    :type value: str
    :return: Key to retrieve the url.
    :rtype: str
    """
    while True:
        key = generate_key()
        result = redis_client.set(f'{URL_KEYS_DOMAIN}:{key}', value, nx=True)
        if result is not None:
            break
    return key


def update_url(key: str, value: str | None) -> str:
    """Update url by passed key and value.

    :param key: Key to find.
    :type key: str
    :param value: Url value to update.
    :type value: str
    :return: New url value or None if no key found.
    :rtype: str | None
    """
    url_value = get_url_value(key)
    if url_value is None:
        raise NotFound('Url not found')

    redis_client.set(f'{URL_KEYS_DOMAIN}:{key}', value)
    return value


def delete_url(key: str):
    """Delete url by passed key.

    :param key: Url key to delete.
    :type key: str
    """
    if get_url_value(key) is None:
        raise NotFound('Url not found')

    return redis_client.delete(f'{URL_KEYS_DOMAIN}:{key}')
# endregion


# region Other functions
def generate_key() -> str:
    """Generate new key.

    :return: New key.
    :rtype: str
    """
    return ''.join(SystemRandom().choice(ascii_lowercase) for _ in range(6))
# endregion


# region CLI functions
@cli.command('new', help='Insert new url.')
@option('-u', '--url', type=STRING, prompt='Enter the url', help='Specify the url.')
def new_url(url: str):
    """Insert new url.

    :param url: Url value.
    :type url: str
    """
    print(f'Insert url: {url}...')
    key = insert_url(url)
    print(f'Done.{linesep}Use the following key to retrieve it: {key}.')


@cli.command('del', help='Delete url by key.')
@option('-k', '--key', type=STRING, prompt='Enter the key', help='Specify the key.')
def del_url(key: str):
    """Delete url by key.

    :param key; Key.
    :type key: str
    """
    print(f'Deleting url: {key}...')
    delete_url(key)
    print('Done.')
# endregion



# coding=utf-8

"""shortipy.services.redis file."""

from flask import Flask
from flask_redis import FlaskRedis

redis_client = FlaskRedis()


def init_app(app: Flask) -> Flask:
    """Initializes the application Redis client.

    :param app: The Flask application instance.
    :type app: Flask
    :return: The Flask application instance.
    :rtype: Flask
    """
    redis_client.init_app(app, decode_responses=True)
    return app



# coding=utf-8

# pylint: disable=invalid-name

"""shortipy.services.config file."""


class Config:  # pylint: disable=too-few-public-methods
    """Class to manage basic configurations."""

    def __init__(self):
        """Config constructor."""
        self.DEBUG = False

        # Flask
        self.SECRET_KEY = None
        self.SESSION_COOKIE_SECURE = True
        self.SESSION_COOKIE_HTTPONLY = True

        # Flask JWT Extended
        self.JWT_SECRET_KEY = None

        # Flask Redis
        self.REDIS_URL = 'redis://127.0.0.1:6379/0'

        # Flask Marshmallow
        self.JSON_SORT_KEYS = False

    def get_dict(self) -> dict:
        """Get Config dict.

        :return: Config dict.
        :rtype: dict
        """
        result = {}
        for key in dir(self):
            if key.isupper():
                result[key] = getattr(self, key)
        return result



# coding=utf-8

"""shortipy.services.hash file."""

from hashlib import sha256

from flask import Flask
from flask_bcrypt import Bcrypt

bcrypt = Bcrypt()


def init_app(app: Flask) -> Flask:
    """Initializes the application hash.

    :param app: The Flask application instance.
    :type app: Flask
    :return: The Flask application instance.
    :rtype: Flask
    """
    bcrypt.init_app(app)
    return app


def normalize_input(value: str | bytes) -> str:
    """Normalize input value.

    :param value: Input value.
    :type value: str | bytes
    :return: Normalized value.
    :rtype: str
    """
    if not value:
        raise Exception('Invalid value')
    if isinstance(value, str):
        value = bytes(value, 'utf-8')
    return sha256(value).hexdigest()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shortipy',
 'shortipy.controllers',
 'shortipy.controllers.api',
 'shortipy.services']

package_data = \
{'': ['*']}

install_requires = \
['flask-bcrypt>=1.0.1,<2.0.0',
 'flask-jwt-extended>=4.4.4,<5.0.0',
 'flask-marshmallow>=0.14.0,<0.15.0',
 'flask-redis>=0.4.0,<0.5.0',
 'flask>=2.2.2,<3.0.0',
 'webargs>=8.2.0,<9.0.0']

setup_kwargs = {
    'name': 'shortipy',
    'version': '1.3.0',
    'description': 'Shortipy is a RESTful Web API, written in Python language and based on the Flask micro-framework, designed to manage shortened links.',
    'long_description': '![shortipy logo](https://github.com/CoffeePerry/shortipy/blob/main/art/shortipy.png?raw=true)\n\n![GitHub release (latest by date)](https://img.shields.io/github/v/release/CoffeePerry/shortipy)\n![PyPI](https://img.shields.io/pypi/v/shortipy?logo=PyPI&logoColor=white)\n![PyPI - Status](https://img.shields.io/pypi/status/shortipy)\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/shortipy?logo=Python&logoColor=white)\n![PyPI - Implementation](https://img.shields.io/pypi/implementation/shortipy)\n\n[![Flask](https://img.shields.io/badge/framework-Flask-black?logo=Flask&logoColor=white)](https://github.com/pallets/flask/)\n\n[![GitHub license](https://img.shields.io/github/license/CoffeePerry/shortipy)](https://github.com/CoffeePerry/shortipy/blob/master/LICENSE)\n\n[![GitHub issues](https://img.shields.io/github/issues/CoffeePerry/shortipy)](https://github.com/CoffeePerry/shortipy/issues)\n\n# shortipy\n\n**Shortipy** is a *RESTful Web API*, written in *Python* language and based on the *Flask* micro-framework, designed to manage shortened links. \n',
    'author': 'Simone Perini',
    'author_email': 'perinisimone98@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/CoffeePerry/shortipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
